use cgmath::Point2;
use galactica_content::{Content, FactionHandle, OutfitHandle, ShipHandle, SystemHandle};
use galactica_playeragent::PlayerAgent;
use galactica_system::data::ShipPersonality;
use galactica_system::phys::{
	ParticleBuilder, PhysSim, PhysSimShipHandle, PhysStepResources, Wrapper,
};
use galactica_util::timing::Timing;
use rand::seq::SliceRandom;
use std::time::Instant;

#[derive(Clone)]
pub struct GameState {
	pub systemsim: PhysSim,
	pub timing: Timing,
	pub start_instant: Instant,
}

unsafe impl Send for GameState {}
unsafe impl Sync for GameState {}

pub struct Game {
	// Core game data
	ct: Content,
	state: GameState,

	// Metadata
	wrapper: Wrapper, // Physics computer
	time_scale: f32,
	last_update: Instant,

	/// Particles to create this frame.
	/// Must be cleared at the start of every frame
	/// TODO: better way to handle this?
	new_particles: Vec<ParticleBuilder>,
}

unsafe impl<'a> Send for Game {}

impl<'a> Game {
	pub fn make_player(&mut self) -> PhysSimShipHandle {
		let player = self.state.systemsim.add_ship(
			&self.ct,
			ShipHandle { index: 0 },
			FactionHandle { index: 0 },
			ShipPersonality::Player,
			Point2 { x: 0.0, y: 0.0 },
		);

		let s = self.state.systemsim.get_ship_mut(&player).unwrap();
		s.data
			.add_outfit(&self.ct.get_outfit(OutfitHandle { index: 0 }));
		s.data
			.add_outfit(&self.ct.get_outfit(OutfitHandle { index: 1 }));
		s.data
			.add_outfit(&self.ct.get_outfit(OutfitHandle { index: 2 }));

		return player;
	}

	pub fn new(ct: Content) -> Self {
		let mut systemsim = PhysSim::new(&ct, SystemHandle { index: 0 });

		let a = systemsim.add_ship(
			&ct,
			ShipHandle { index: 0 },
			FactionHandle { index: 1 },
			ShipPersonality::Point,
			Point2 { x: 100.0, y: 0.0 },
		);

		let s = systemsim.get_ship_mut(&a).unwrap();
		s.data.add_outfit(ct.get_outfit(OutfitHandle { index: 0 }));
		s.data.add_outfit(&ct.get_outfit(OutfitHandle { index: 1 }));
		s.data.add_outfit(&ct.get_outfit(OutfitHandle { index: 2 }));

		let a = systemsim.add_ship(
			&ct,
			ShipHandle { index: 0 },
			FactionHandle { index: 0 },
			ShipPersonality::Point,
			Point2 { x: 0.0, y: 120.0 },
		);

		let s = systemsim.get_ship_mut(&a).unwrap();
		s.data.add_outfit(ct.get_outfit(OutfitHandle { index: 0 }));
		s.data.add_outfit(&ct.get_outfit(OutfitHandle { index: 1 }));
		s.data.add_outfit(&ct.get_outfit(OutfitHandle { index: 2 }));

		let state = GameState {
			systemsim,
			timing: Timing::new(),
			start_instant: Instant::now(),
		};

		Game {
			ct,
			state,

			wrapper: Wrapper::new(),
			last_update: Instant::now(),
			time_scale: 1.0,
			new_particles: Vec::new(),
		}
	}

	pub fn update_player_controls(&mut self, player: &mut PlayerAgent) {
		self.state
			.systemsim
			.update_player_controls(&self.ct, player)
	}

	pub fn get_state(&self) -> &GameState {
		&self.state
	}

	pub fn get_particles(&self) -> &Vec<ParticleBuilder> {
		&self.new_particles
	}

	pub fn update(&mut self) {
		self.state.timing.start_frame();
		let t: f32 = self.last_update.elapsed().as_secs_f32() * self.time_scale;
		self.new_particles.clear();

		self.state.systemsim.step(PhysStepResources {
			ct: &self.ct,
			particles: &mut self.new_particles,
			timing: &mut self.state.timing,
			wrapper: &mut self.wrapper,
			t,
		});

		self.last_update = Instant::now();
		self.new_particles.shuffle(&mut rand::thread_rng());
		self.state.timing.mark_frame();
	}
}
