mod game;

use anyhow::{bail, Result};
use cgmath::Point2;
use galactica_content::{Content, SystemHandle};
use galactica_playeragent::{PlayerAgent, PlayerStatus};
use galactica_render::RenderInput;
use galactica_system::{
	data::ShipState,
	phys::{util, PhysSimShipHandle},
};
use galactica_util::constants::ASSET_CACHE;
use std::{
	fs,
	path::{Path, PathBuf},
};
use winit::{
	event::{Event, KeyboardInput, WindowEvent},
	event_loop::{ControlFlow, EventLoop},
	window::WindowBuilder,
};

fn main() -> Result<()> {
	let cache_dir = Path::new(ASSET_CACHE);
	fs::create_dir_all(cache_dir)?;

	let atlas_index = cache_dir.join("spriteatlas.toml");

	if !atlas_index.exists() {
		bail!("Could not find sprite atlas!");
	}

	// TODO: pretty error if missing
	let content = Content::load_dir(
		PathBuf::from("./content"),
		PathBuf::from("./assets"),
		atlas_index.clone(),
	)?;

	let event_loop = EventLoop::new();
	let window = WindowBuilder::new().build(&event_loop).unwrap();
	let mut gpu = pollster::block_on(galactica_render::GPUState::new(window, &content))?;
	gpu.init(&content);

	let mut game = game::Game::new(content.clone());
	let p = game.make_player();

	let mut player = PlayerAgent::new(p.0);
	player.set_camera_aspect(
		gpu.window().inner_size().width as f32 / gpu.window().inner_size().height as f32,
	);

	event_loop.run(move |event, _, control_flow| {
		match event {
			Event::RedrawRequested(window_id) if window_id == gpu.window().id() => {
				let render_input = RenderInput {
					camera_pos: player.camera.pos,
					camera_zoom: player.camera.zoom,
					current_time: game.get_state().start_instant.elapsed().as_secs_f32(),
					ct: &content,
					systemsim: &game.get_state().systemsim,
					particles: game.get_particles(),
					player: &player,
					current_system: SystemHandle { index: 0 },
					timing: game.get_state().timing.clone(),
				};

				match gpu.render(render_input) {
					Ok(_) => {}
					Err(wgpu::SurfaceError::Lost) => gpu.resize(&content),
					Err(wgpu::SurfaceError::OutOfMemory) => *control_flow = ControlFlow::Exit,
					// All other errors (Outdated, Timeout) should be resolved by the next frame
					Err(e) => eprintln!("{:?}", e),
				}
			}

			Event::MainEventsCleared => {
				game.update_player_controls(&mut player);
				game.update();

				// TODO: clean up
				let player_status = {
					let pos = {
						let o = &game
							.get_state()
							.systemsim
							.get_ship(&PhysSimShipHandle(player.ship.unwrap()));
						if let Some(o) = o {
							match o.data.get_state() {
								ShipState::Collapsing { .. } | ShipState::Flying => {
									let r =
										&game.get_state().systemsim.get_rigid_body(o.rigid_body);
									if let Some(r) = r {
										Some(util::rigidbody_position(r))
									} else {
										None
									}
								}
								ShipState::Landed { target } => {
									let b = content.get_system_object(*target);
									Some(Point2 {
										x: b.pos.x,
										y: b.pos.y,
									})
								}
							}
						} else {
							None
						}
					};

					PlayerStatus { pos }
				};

				// This must be updated BEFORE rendering!
				player.step(&content, player_status);

				player.input.clear_inputs();
				gpu.window().request_redraw();
			}

			Event::WindowEvent {
				ref event,
				window_id,
			} if window_id == gpu.window().id() => match event {
				WindowEvent::Focused(_state) => {
					//game.set_paused(!state);
				}
				WindowEvent::CloseRequested => *control_flow = ControlFlow::Exit,
				WindowEvent::KeyboardInput {
					input:
						KeyboardInput {
							state,
							virtual_keycode: Some(key),
							..
						},
					..
				} => {
					player.process_key(state, key);
				}
				WindowEvent::MouseInput { state, button, .. } => {
					player.process_click(state, button);
				}
				WindowEvent::MouseWheel { delta, phase, .. } => {
					player.process_scroll(delta, phase);
				}
				WindowEvent::Resized(_) => {
					gpu.resize(&content);
					player.set_camera_aspect(
						gpu.window().inner_size().width as f32
							/ gpu.window().inner_size().height as f32,
					);
				}
				WindowEvent::ScaleFactorChanged { .. } => {
					gpu.resize(&content);
					player.set_camera_aspect(
						gpu.window().inner_size().width as f32
							/ gpu.window().inner_size().height as f32,
					);
				}
				_ => {}
			},
			_ => {}
		}
	});
}
