use cgmath::{Deg, EuclideanSpace, InnerSpace, Matrix2, Rad, Vector2, Zero};
use galactica_content::{CollapseEvent, Ship};
use nalgebra::point;
use rand::{rngs::ThreadRng, Rng};
use rapier2d::{dynamics::RigidBody, geometry::Collider};

use super::super::{util, ParticleBuilder, PhysStepResources};
use crate::data::ShipData;

#[derive(Debug, Clone)]
pub(super) struct ShipCollapseSequence {
	rng: ThreadRng,

	/// The elapsed collapse duration when step()
	/// was last called
	last_call: f32,
}

impl ShipCollapseSequence {
	pub(super) fn new() -> Self {
		Self {
			rng: rand::thread_rng(),
			last_call: 0.0,
		}
	}

	/// Pick a random points inside a ship's collider
	fn random_in_ship(&mut self, ship: &Ship, collider: &Collider) -> Vector2<f32> {
		let mut y = 0.0;
		let mut x = 0.0;
		let mut a = false;
		while !a {
			y = self.rng.gen_range(-1.0..=1.0) * ship.size / 2.0;
			x = self.rng.gen_range(-1.0..=1.0) * ship.size * ship.sprite.aspect / 2.0;
			a = collider.shape().contains_local_point(&point![x, y]);
		}
		Vector2 { x, y }
	}

	/// Step this sequence `t` seconds
	pub(super) fn step(
		&mut self,
		res: &mut PhysStepResources,
		ship_data: &ShipData,
		rigid_body: &mut RigidBody,
		collider: &mut Collider,
	) {
		let ship_content = res.ct.get_ship(ship_data.get_content());
		let ship_pos = util::rigidbody_position(rigid_body);
		let ship_rot = util::rigidbody_rotation(rigid_body);
		let ship_ang = ship_rot.angle(Vector2 { x: 1.0, y: 0.0 });

		let (total, elapsed) = ship_data.get_state().collapse_state().unwrap();
		// How much time has passed since step() was last called
		let delta = elapsed - self.last_call;

		// The fraction of this collapse sequence that has been played
		let frac_done = elapsed / total;

		// TODO: slight random offset for event particles

		// Trigger collapse events
		for event in &ship_content.collapse.events {
			match event {
				CollapseEvent::Effect(event) => {
					if (event.time > self.last_call && event.time <= elapsed)
						|| (event.time == 0.0 && self.last_call == 0.0)
					// ^^ Don't miss events scheduled at the very start of the sequence!
					{
						for spawner in &event.effects {
							let effect = res.ct.get_effect(spawner.effect);

							for _ in 0..spawner.count as usize {
								let pos = if let Some(pos) = spawner.pos {
									pos.to_vec()
								} else {
									self.random_in_ship(ship_content, collider)
								};
								let pos = ship_pos
									+ (Matrix2::from_angle(-ship_ang - Rad::from(Deg(90.0))) * pos);

								let velocity = rigid_body.velocity_at_point(&point![pos.x, pos.y]);

								res.particles.push(ParticleBuilder::from_content(
									effect,
									pos,
									Rad::zero(),
									Vector2 {
										x: velocity.x,
										y: velocity.y,
									},
									Vector2::zero(),
								))
							}
						}
					}
				}
			}
		}

		// Create collapse effects
		for spawner in &ship_content.collapse.effects {
			let effect = res.ct.get_effect(spawner.effect);

			// Probability of adding a particle this frame.
			// The area of this function over [0, 1] should be 1.
			let pdf = |x: f32| {
				let f = 0.2;
				let y = if x < (1.0 - f) {
					let x = x / (1.0 - f);
					(x * x + 0.2) * 1.8 - f
				} else {
					1.0
				};
				return y;
			};

			// Notice that we don't use res.t here, since ship state is updated earlier
			// ( through self.data.step() )
			let p_add = (delta / total) * pdf(frac_done) * spawner.count;

			if self.rng.gen_range(0.0..=1.0) <= p_add {
				let pos = if let Some(pos) = spawner.pos {
					pos.to_vec()
				} else {
					self.random_in_ship(ship_content, collider)
				};

				// Position, adjusted for ship rotation
				let pos = ship_pos + Matrix2::from_angle(-ship_ang - Rad::from(Deg(90.0))) * pos;
				let vel = rigid_body.velocity_at_point(&point![pos.x, pos.y]);

				res.particles.push(ParticleBuilder {
					sprite: effect.sprite,
					pos,
					velocity: Vector2 { x: vel.x, y: vel.y },
					angle: Rad::zero(),
					angvel: Rad::zero(),
					lifetime: effect.lifetime,
					size: effect.size,
					fade: 0.0,
				});
			}
		}

		self.last_call = elapsed;
	}
}
