use cgmath::{Deg, EuclideanSpace, InnerSpace, Matrix2, Point2, Rad, Vector2, Zero};
use galactica_content::{
	Content, FactionHandle, GunPoint, OutfitHandle, ProjectileCollider, Relationship, ShipHandle,
	SystemHandle, SystemObjectHandle,
};
use galactica_playeragent::PlayerAgent;
use nalgebra::{point, vector};
use rand::Rng;
use rapier2d::{
	dynamics::{RigidBody, RigidBodyBuilder, RigidBodyHandle, RigidBodySet},
	geometry::{ColliderBuilder, ColliderHandle, ColliderSet},
	pipeline::ActiveEvents,
};
use std::{collections::HashMap, f32::consts::PI};

use crate::data::{ShipPersonality, ShipState};

use super::{
	controller::ShipController,
	objects::{PhysProjectile, PhysSimShip},
	util, ParticleBuilder, PhysStepResources,
};

// TODO: replace with a more generic handle
/// A handle for a ship in this simulation
/// This lets other crates reference ships
/// without including `rapier2d`.
pub struct PhysSimShipHandle(pub ColliderHandle);

/// Manages the physics state of one system
#[derive(Clone)]
pub struct PhysSim {
	/// The system this sim is attached to
	_system: SystemHandle,

	rigid_body_set: RigidBodySet,
	collider_set: ColliderSet,

	projectiles: HashMap<ColliderHandle, PhysProjectile>,
	ships: HashMap<ColliderHandle, PhysSimShip>,
	ship_behaviors: HashMap<ColliderHandle, ShipController>,
}

// Private methods
impl<'a> PhysSim {
	fn remove_projectile(
		&mut self,
		res: &mut PhysStepResources,
		c: ColliderHandle,
	) -> Option<(RigidBody, PhysProjectile)> {
		let p = match self.projectiles.remove(&c) {
			Some(p) => p,
			None => return None,
		};
		let r = self
			.rigid_body_set
			.remove(
				p.rigid_body,
				&mut res.wrapper.im,
				&mut self.collider_set,
				&mut res.wrapper.ij,
				&mut res.wrapper.mj,
				true,
			)
			.unwrap();

		return Some((r, p));
	}

	fn land_ship(&mut self, collider: ColliderHandle, target: SystemObjectHandle) {
		let ship = self.ships.get_mut(&collider).unwrap();
		self.rigid_body_set
			.get_mut(ship.rigid_body)
			.unwrap()
			.set_enabled(false);
		self.collider_set
			.get_mut(ship.collider)
			.unwrap()
			.set_enabled(false);

		ship.data.land_on(target);
	}

	fn unland_ship(&mut self, ct: &Content, collider: ColliderHandle) {
		let ship = self.ships.get_mut(&collider).unwrap();
		let obj = ship.data.get_state().landed_on().unwrap();
		let obj = ct.get_system_object(obj);
		ship.data.unland();

		self.rigid_body_set
			.get_mut(ship.rigid_body)
			.unwrap()
			.set_position(point![obj.pos.x, obj.pos.y].into(), true);
		self.rigid_body_set
			.get_mut(ship.rigid_body)
			.unwrap()
			.set_enabled(true);
		self.collider_set
			.get_mut(ship.collider)
			.unwrap()
			.set_enabled(true);
	}

	fn remove_ship(&mut self, res: &mut PhysStepResources, colliderhandle: ColliderHandle) {
		let ship = match self.ships.get(&colliderhandle) {
			None => return,
			Some(s) => s,
		};

		self.rigid_body_set.remove(
			ship.rigid_body,
			&mut res.wrapper.im,
			&mut self.collider_set,
			&mut res.wrapper.ij,
			&mut res.wrapper.mj,
			true,
		);
		self.ships.remove(&colliderhandle).unwrap();
		self.ship_behaviors.remove(&colliderhandle);
	}

	fn collide_projectile_ship(
		&mut self,
		res: &mut PhysStepResources,
		projectile_h: ColliderHandle,
		ship_h: ColliderHandle,
	) {
		let projectile = self.projectiles.get(&projectile_h);
		let ship = self.ships.get_mut(&ship_h);
		if projectile.is_none() || ship.is_none() {
			return;
		}
		let projectile = projectile.unwrap();
		let ship = ship.unwrap();

		let f = res.ct.get_faction(projectile.faction);
		let r = f.relationships.get(&ship.data.get_faction()).unwrap();
		let destory_projectile = match r {
			Relationship::Hostile => {
				ship.data.apply_damage(res.ct, projectile.content.damage);
				true
			}
			_ => false,
		};

		if destory_projectile {
			let pr = self.rigid_body_set.get(projectile.rigid_body).unwrap();
			let v = util::rigidbody_velocity(pr).normalize() * projectile.content.force;
			let pos = util::rigidbody_position(pr);
			let _ = pr;

			let r = self.rigid_body_set.get_mut(ship.rigid_body).unwrap();
			r.apply_impulse_at_point(vector![v.x, v.y], point![pos.x, pos.y], true);

			// Borrow again, we can only have one at a time
			let pr = self.rigid_body_set.get(projectile.rigid_body).unwrap();
			let pos = util::rigidbody_position(pr);
			let angle = util::rigidbody_rotation(pr).angle(Vector2 { x: 1.0, y: 0.0 });

			match &projectile.content.impact_effect {
				None => {}
				Some(x) => {
					let effect = res.ct.get_effect(*x);
					let r = ship.rigid_body;
					let sr = self.get_rigid_body(r).unwrap();
					let parent_velocity = util::rigidbody_velocity(pr);
					let target_velocity =
						sr.velocity_at_point(&nalgebra::Point2::new(pos.x, pos.y));

					res.particles.push(ParticleBuilder::from_content(
						effect,
						pos,
						-angle,
						parent_velocity,
						Vector2 {
							x: target_velocity.x,
							y: target_velocity.y,
						},
					));
				}
			};

			self.remove_projectile(res, projectile.collider);
		}
	}
}

// Public methods
impl PhysSim {
	/// Create a new physics system
	pub fn new(_ct: &Content, system: SystemHandle) -> Self {
		Self {
			_system: system,
			rigid_body_set: RigidBodySet::new(),
			collider_set: ColliderSet::new(),
			projectiles: HashMap::new(),
			ships: HashMap::new(),
			ship_behaviors: HashMap::new(),
		}
	}

	/// Add a ship to this physics system
	pub fn add_ship(
		&mut self,
		ct: &Content,
		handle: ShipHandle,
		faction: FactionHandle,
		personality: ShipPersonality,
		position: Point2<f32>,
	) -> PhysSimShipHandle {
		let ship_content = ct.get_ship(handle);
		let cl = ColliderBuilder::convex_decomposition(
			&ship_content.collision.points[..],
			&ship_content.collision.indices[..],
		)
		// Rotate collider to match sprite
		// (Collider starts pointing east, sprite starts pointing north.)
		.rotation(PI / -2.0)
		.mass(ship_content.mass);
		// TODO: only build colliders once

		let rb = RigidBodyBuilder::dynamic()
			.angular_damping(ship_content.angular_drag)
			.linear_damping(ship_content.linear_drag)
			.translation(vector![position.x, position.y])
			.can_sleep(false);

		let ridid_body = self.rigid_body_set.insert(rb.build());
		let collider =
			self.collider_set
				.insert_with_parent(cl.build(), ridid_body, &mut self.rigid_body_set);

		self.ship_behaviors.insert(
			collider,
			match personality {
				ShipPersonality::Dummy | ShipPersonality::Player => ShipController::new_null(),
				ShipPersonality::Point => ShipController::new_point(),
			},
		);

		self.ships.insert(
			collider,
			PhysSimShip::new(ct, handle, personality, faction, ridid_body, collider),
		);

		return PhysSimShipHandle(collider);
	}

	/// Update a player ship's controls
	pub fn update_player_controls(&mut self, ct: &Content, player: &PlayerAgent) {
		if player.ship.is_none() {
			return;
		}
		let ship_object = self.ships.get_mut(&player.ship.unwrap());
		if let Some(ship_object) = ship_object {
			ship_object.controls.guns = player.input.pressed_guns();
			ship_object.controls.left = player.input.pressed_left();
			ship_object.controls.right = player.input.pressed_right();
			ship_object.controls.thrust = player.input.pressed_thrust();

			if player.input.pressed_land() {
				match ship_object.data.get_state() {
					ShipState::Flying => {
						self.land_ship(
							player.ship.unwrap(),
							player.selection.get_planet().unwrap(),
						);
					}
					ShipState::Landed { .. } => {
						self.unland_ship(ct, player.ship.unwrap());
					}
					_ => {}
				}
			}
		}
	}

	/// Run ship updates:
	/// - updates ship controls (runs behaviors)
	/// - applies ship controls
	/// - creates projectiles
	fn step_ships(&mut self, res: &mut PhysStepResources) {
		// We can't apply these right away since self is borrowed
		// by the iterator
		// TODO: don't allocate!
		let mut projectiles = Vec::new();
		let mut to_remove = Vec::new();
		// Again, borrow checker hack. TODO: fix
		let keys: Vec<_> = self.ships.keys().map(|c| *c).collect();
		for collider in keys {
			// Borrow immutably for now...
			// (required to compute controls)
			let ship = self.ships.get(&collider).unwrap();
			if ship.data.get_state().should_be_removed() {
				to_remove.push(collider);
				continue;
			}

			match ship.data.get_state() {
				ShipState::Landed { .. } => {}

				ShipState::Collapsing { .. } => {
					let ship = self.ships.get_mut(&collider).unwrap();
					ship.step(
						res,
						&mut self.rigid_body_set[ship.rigid_body],
						&mut self.collider_set[ship.collider],
					);
				}

				ShipState::Flying => {
					// Compute new controls
					// This is why we borrow immutably first
					let controls;
					let b = self.ship_behaviors.get_mut(&collider).unwrap();
					controls =
						b.update_controls(&res, &self.rigid_body_set, &self.ships, ship.collider);

					// Re-borrow mutably to apply changes
					let ship = self.ships.get_mut(&collider).unwrap();

					if let Some(controls) = controls {
						ship.controls = controls;
					}
					ship.step(
						res,
						&mut self.rigid_body_set[ship.rigid_body],
						&mut self.collider_set[ship.collider],
					);

					// If we're firing, try to fire each gun
					if ship.controls.guns {
						// TODO: don't allocate here. This is a hack to satisfy the borrow checker,
						// convert this to a refcell or do the replace dance.
						let pairs: Vec<(GunPoint, Option<OutfitHandle>)> = ship
							.data
							.get_outfits()
							.iter_gun_points()
							.map(|(p, o)| (p.clone(), o.clone()))
							.collect();

						for (gun, outfit) in pairs {
							if ship.data.fire_gun(res.ct, &gun) {
								projectiles.push((ship.collider, gun.clone(), outfit.unwrap()));
							}
						}
					}
				}
			}
		}

		// Remove ships that don't exist
		for c in to_remove {
			self.remove_ship(res, c);
		}

		// Create projectiles
		for (collider, gun_point, outfit) in projectiles {
			let mut rng = rand::thread_rng();

			let ship = self.ships.get(&collider).unwrap();
			let rigid_body = self.get_rigid_body(ship.rigid_body).unwrap();
			let ship_pos = util::rigidbody_position(rigid_body);
			let ship_rot = util::rigidbody_rotation(rigid_body);
			let ship_vel = util::rigidbody_velocity(rigid_body);
			let ship_ang = ship_rot.angle(Vector2 { x: 0.0, y: 1.0 });

			let pos = ship_pos + (Matrix2::from_angle(-ship_ang) * gun_point.pos.to_vec());

			let outfit = res.ct.get_outfit(outfit);
			let outfit = outfit.gun.as_ref().unwrap();

			let spread: Rad<f32> =
				Deg(rng.gen_range(-outfit.projectile.angle_rng.0..=outfit.projectile.angle_rng.0))
					.into();

			let vel = ship_vel
				+ (Matrix2::from_angle(-ship_ang + spread)
					* Vector2 {
						x: 0.0,
						y: outfit.projectile.speed
							+ rng.gen_range(
								-outfit.projectile.speed_rng..=outfit.projectile.speed_rng,
							),
					});

			let rigid_body = RigidBodyBuilder::kinematic_velocity_based()
				.translation(vector![pos.x, pos.y])
				.rotation(-ship_ang.0)
				.linvel(vector![vel.x, vel.y])
				.build();

			let collider = match &outfit.projectile.collider {
				ProjectileCollider::Ball(b) => ColliderBuilder::ball(b.radius)
					.sensor(true)
					.active_events(ActiveEvents::COLLISION_EVENTS)
					.build(),
			};

			let rigid_body = self.rigid_body_set.insert(rigid_body);
			let collider = self.collider_set.insert_with_parent(
				collider,
				rigid_body,
				&mut self.rigid_body_set,
			);

			self.projectiles.insert(
				collider.clone(),
				PhysProjectile::new(
					outfit.projectile.clone(),
					rigid_body,
					ship.data.get_faction(),
					collider,
				),
			);
		}
	}

	/// Step this physics system by `t` seconds
	pub fn step(&mut self, mut res: PhysStepResources) {
		res.timing.start_physics_ships();
		self.step_ships(&mut res);
		res.timing.mark_physics_ships();

		res.timing.start_physics_sim();

		// Update physics
		res.wrapper
			.step(res.t, &mut self.rigid_body_set, &mut self.collider_set);

		// Handle collision events
		while let Ok(event) = &res.wrapper.collision_queue.try_recv() {
			if event.started() {
				let a = event.collider1();
				let b = event.collider2();

				// If projectiles are part of this collision, make sure
				// `a` is one of them.
				let (a, b) = if self.projectiles.contains_key(&b) {
					(b, a)
				} else {
					(a, b)
				};

				let p = self.projectiles.get(&a);
				let s = self.ships.get_mut(&b);
				if p.is_none() || s.is_none() {
					continue;
				}
				self.collide_projectile_ship(&mut res, a, b);
			}
		}

		// Delete projectiles
		let mut to_remove = Vec::new();
		for (c, p) in &mut self.projectiles {
			p.tick(res.t);
			if p.is_expired() {
				to_remove.push(*c);
			}
		}

		let mut rng = rand::thread_rng();
		for c in to_remove {
			let (pr, p) = self.remove_projectile(&mut res, c).unwrap();

			match &p.content.expire_effect {
				None => {}
				Some(x) => {
					let x = res.ct.get_effect(*x);
					let pos = util::rigidbody_position(&pr);
					let vel = util::rigidbody_velocity(&pr);
					let angle = util::rigidbody_rotation(&pr).angle(Vector2 { x: 1.0, y: 0.0 });

					let velocity = {
						let a = rng
							.gen_range(-x.velocity_scale_parent_rng..=x.velocity_scale_parent_rng);

						let velocity = (x.velocity_scale_parent + a) * vel;

						velocity
					};

					res.particles.push(ParticleBuilder::from_content(
						x,
						pos,
						-angle,
						velocity,
						Vector2::zero(),
					));
				}
			};
		}

		res.timing.mark_physics_sim();
	}
}

// Public getters
impl PhysSim {
	/// Get a ship physics object
	pub fn get_ship(&self, ship: &PhysSimShipHandle) -> Option<&PhysSimShip> {
		self.ships.get(&ship.0)
	}

	/// Get a ship physics object
	pub fn get_ship_mut(&mut self, ship: &PhysSimShipHandle) -> Option<&mut PhysSimShip> {
		self.ships.get_mut(&ship.0)
	}

	/// Get a rigid body from a handle
	pub fn get_rigid_body(&self, r: RigidBodyHandle) -> Option<&RigidBody> {
		self.rigid_body_set.get(r)
	}

	/// Get a rigid body from a handle
	pub fn get_rigid_body_mut(&mut self, r: RigidBodyHandle) -> Option<&mut RigidBody> {
		self.rigid_body_set.get_mut(r)
	}

	/// Iterate over all ships in this physics system
	pub fn iter_ship_body(&self) -> impl Iterator<Item = (&PhysSimShip, &RigidBody)> + '_ {
		self.ships
			.values()
			.map(|x| (x, self.rigid_body_set.get(x.rigid_body).unwrap()))
	}

	/// Iterate over all ships in this physics system
	pub fn iter_ships(&self) -> impl Iterator<Item = &PhysSimShip> + '_ {
		self.ships.values()
	}

	/// Iterate over all ships in this physics system
	pub fn iter_projectiles(&self) -> impl Iterator<Item = &PhysProjectile> + '_ {
		self.projectiles.values()
	}
}
